"""
gargoyle.templatetags.gargoyle_helpers
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

:copyright: (c) 2010 DISQUS.
:license: Apache License 2.0, see LICENSE for more details.
"""
from __future__ import absolute_import, division, print_function, unicode_literals

from django import template

register = template.Library()


@register.filter
def render_field(field, value=None):
    return field.render(value)


@register.filter
def sort_by_key(field, currently):
    is_negative = currently.find('-') is 0
    current_field = currently.lstrip('-')

    if current_field == field and is_negative:
        return field
    elif current_field == field:
        return '-' + field
    else:
        return field


@register.filter
def sort_field(sort_string):
    return sort_string.lstrip('-')



"""
gargoyle.nexus_modules
~~~~~~~~~~~~~~~~~~~~~~

:copyright: (c) 2010 DISQUS.
:license: Apache License 2.0, see LICENSE for more details.
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import logging
from functools import wraps

import nexus
from django.conf import settings
from django.conf.urls import url
from django.http import HttpResponse, HttpResponseNotFound
from django.utils import six

from gargoyle import gargoyle, signals
from gargoyle.conditions import ValidationError
from gargoyle.helpers import dumps
from gargoyle.models import DISABLED, Switch

logger = logging.getLogger('gargoyle.switches')


class GargoyleException(Exception):
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return self.message


def json_view(func):
    "Decorator to make JSON views simpler"

    @wraps(func)
    def wrapper(self, request, *args, **kwargs):
        try:
            response = {
                "success": True,
                "data": func(self, request, *args, **kwargs)
            }
        except GargoyleException as exc:
            response = {
                "success": False,
                "data": exc.message
            }
        except Switch.DoesNotExist:
            response = {
                "success": False,
                "data": "Switch cannot be found"
            }
        except ValidationError as e:
            response = {
                "success": False,
                "data": u','.join(map(six.text_type, e.messages)),
            }
        except Exception:
            if settings.DEBUG:
                import traceback
                traceback.print_exc()
            raise
        return HttpResponse(dumps(response), content_type="application/json")

    return wrapper


class GargoyleModule(nexus.NexusModule):
    home_url = 'index'
    name = 'gargoyle'

    def get_title(self):
        return 'Gargoyle'

    def get_urls(self):
        return [
            url(r'^add/$', self.as_view(self.add), name='add'),
            url(r'^update/$', self.as_view(self.update), name='update'),
            url(r'^delete/$', self.as_view(self.delete), name='delete'),
            url(r'^status/$', self.as_view(self.status), name='status'),
            url(r'^conditions/add/$', self.as_view(self.add_condition), name='add-condition'),
            url(r'^conditions/remove/$', self.as_view(self.remove_condition), name='remove-condition'),
            url(r'^$', self.as_view(self.index), name='index'),
        ]

    def render_on_dashboard(self, request):
        active_switches_count = Switch.objects.exclude(status=DISABLED).count()

        switches = list(Switch.objects.exclude(status=DISABLED).order_by("date_created")[:5])

        return self.render_to_string('gargoyle/nexus/dashboard.html', {
            'switches': switches,
            'active_switches_count': active_switches_count,
        })

    def index(self, request):
        sort_by = request.GET.get('by', '-date_modified')

        if sort_by not in self.valid_sort_orders:
            return HttpResponseNotFound('Invalid sort order.')

        switches = list(Switch.objects.all().order_by(sort_by))

        return self.render_to_response("gargoyle/index.html", {
            "switches": [s.to_dict(gargoyle) for s in switches],
            "all_conditions": list(gargoyle.get_all_conditions()),
            "sorted_by": sort_by
        }, request)

    @json_view
    def add(self, request):
        key = request.POST.get("key")

        if not key:
            raise GargoyleException("Key cannot be empty")

        if len(key) > 64:
            raise GargoyleException("Key must be less than or equal to 64 characters in length")

        label = request.POST.get("name", "").strip()

        if len(label) > 64:
            raise GargoyleException("Name must be less than or equal to 64 characters in length")

        switch, created = Switch.objects.get_or_create(
            key=key,
            defaults=dict(
                label=label or None,
                description=request.POST.get("desc")
            )
        )

        if not created:
            raise GargoyleException("Switch with key %s already exists" % key)

        logger.info('Switch %r added (%%s)' % switch.key,
                    ', '.join('%s=%r' % (k, getattr(switch, k)) for k in sorted(('key', 'label', 'description', ))))

        signals.switch_added.send(
            sender=self,
            request=request,
            switch=switch,
        )

        return switch.to_dict(gargoyle)

    @json_view
    def update(self, request):
        switch = Switch.objects.get(key=request.POST.get("curkey"))

        key = request.POST.get("key")

        if len(key) > 64:
            raise GargoyleException("Key must be less than or equal to 64 characters in length")

        label = request.POST.get("name", "")

        if len(label) > 64:
            raise GargoyleException("Name must be less than or equal to 64 characters in length")

        values = dict(
            label=label,
            key=key,
            description=request.POST.get("desc"),
        )

        changes = {}
        for attribute, value in six.iteritems(values):
            new_value = getattr(switch, attribute)
            if new_value != value:
                changes[attribute] = (value, new_value)

        if changes:
            if switch.key != key:
                switch.delete()
                switch.key = key

            switch.label = label
            switch.description = request.POST.get("desc")
            switch.save()

            logger.info('Switch %r updated %%s' % switch.key,
                        ', '.join('%s=%r->%r' % (k, v[0], v[1]) for k, v in sorted(six.iteritems(changes))))

            signals.switch_updated.send(
                sender=self,
                request=request,
                switch=switch,
                changes=changes,
            )

        return switch.to_dict(gargoyle)

    @json_view
    def status(self, request):
        switch = Switch.objects.get(key=request.POST.get("key"))

        try:
            status = int(request.POST.get("status"))
        except ValueError:
            raise GargoyleException("Status must be integer")

        old_status = switch.status
        old_status_label = switch.get_status_display()

        if switch.status != status:
            switch.status = status
            switch.save()

            logger.info('Switch %r updated (status=%%s->%%s)' % switch.key,
                        old_status_label, switch.get_status_display())

            signals.switch_status_updated.send(
                sender=self,
                request=request,
                switch=switch,
                old_status=old_status,
                status=status,
            )

        return switch.to_dict(gargoyle)

    @json_view
    def delete(self, request):
        switch = Switch.objects.get(key=request.POST.get("key"))
        switch.delete()

        logger.info('Switch %r removed' % switch.key)

        signals.switch_deleted.send(
            sender=self,
            request=request,
            switch=switch,
        )

        return {}

    @json_view
    def add_condition(self, request):
        key = request.POST.get("key")
        condition_set_id = request.POST.get("id")
        field_name = request.POST.get("field")
        exclude = int(request.POST.get("exclude") or 0)

        if not all([key, condition_set_id, field_name]):
            raise GargoyleException("Fields cannot be empty")

        field = gargoyle.get_condition_set_by_id(condition_set_id).fields[field_name]
        value = field.validate(request.POST)

        switch = gargoyle[key]
        switch.add_condition(condition_set_id, field_name, value, exclude=exclude)

        logger.info('Condition added to %r (%r, %s=%r, exclude=%r)' % (switch.key,
                    condition_set_id, field_name, value, bool(exclude)))

        signals.switch_condition_added.send(
            sender=self,
            request=request,
            switch=switch,
            condition={
                'condition_set_id': condition_set_id,
                'field_name': field_name,
                'value': value,
            },
        )

        return switch.to_dict(gargoyle)

    @json_view
    def remove_condition(self, request):
        key = request.POST.get("key")
        condition_set_id = request.POST.get("id")
        field_name = request.POST.get("field")
        value = request.POST.get("value")

        if not all([key, condition_set_id, field_name]):
            raise GargoyleException("Fields cannot be empty")

        switch = gargoyle[key]
        switch.remove_condition(condition_set_id, field_name, value)

        logger.info('Condition removed from %r (%r, %s=%r)' % (switch.key,
                    condition_set_id, field_name, value))

        signals.switch_condition_removed.send(
            sender=self,
            request=request,
            switch=switch,
            condition={
                'condition_set_id': condition_set_id,
                'field_name': field_name,
                'value': value,
            },
        )

        return switch.to_dict(gargoyle)

    @property
    def valid_sort_orders(self):
        fields = ['label', 'date_created', 'date_modified']
        return fields + ['-' + f for f in fields]


nexus.site.register(GargoyleModule, 'gargoyle')



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import django.utils.timezone
import jsonfield.fields
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Switch',
            fields=[
                ('key', models.CharField(primary_key=True, serialize=False, max_length=64)),
                ('value', jsonfield.fields.JSONField()),
                ('label', models.CharField(max_length=64, null=True)),
                ('date_created', models.DateTimeField(default=django.utils.timezone.now)),
                ('date_modified', models.DateTimeField(auto_now=True)),
                ('description', models.TextField(null=True)),
                ('status', models.PositiveSmallIntegerField(
                    default=1,
                    choices=[(1, b'Disabled'), (2, b'Selective'), (3, b'Global'), (4, b'Inherit')]
                )),
            ],
            options={
                'verbose_name_plural': 'switches',
                'verbose_name': 'switch',
                'permissions': (('can_view', 'Can view'),),
            },
        ),
    ]



from __future__ import absolute_import, division, print_function, unicode_literals

from django.core.management.base import BaseCommand
from django.utils import six

from gargoyle.models import Switch


class Command(BaseCommand):
    help = 'Removes the specified gargoyle switch.'

    def add_arguments(self, parser):
        parser.add_argument('switch_name', type=six.text_type)

    def handle(self, *args, **options):
        Switch.objects.filter(key=options['switch_name']).delete()



from __future__ import absolute_import, division, print_function, unicode_literals

from django.apps import AppConfig


class GargoyleAppConfig(AppConfig):
    name = 'gargoyle'
    verbose_name = 'Gargoyle'

    def ready(self):
        self.module.autodiscover()



"""
gargoyle.testutils
~~~~~~~~~~~~~~~~~~

:copyright: (c) 2010 DISQUS.
:license: Apache License 2.0, see LICENSE for more details.
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import inspect
import sys
import unittest

from gargoyle import gargoyle
from .compat import ContextDecorator


class TestCaseContextDecorator(ContextDecorator):
    """
    ContextDecorator subclass that allows the sane decoration of TestCase classes by wrapping
    from setUpClass to tearDownClass
    """
    def __call__(self, decorable):
        if inspect.isclass(decorable):
            if not issubclass(decorable, unittest.TestCase):
                raise ValueError("Only supports the wrapping of unittest.TestCase classes")

            klass = decorable

            orig_setUpClass = klass.setUpClass
            orig_tearDownClass = klass.tearDownClass

            def setUpClass(cls):
                self.__enter__()
                try:
                    if orig_setUpClass is not None:
                        super  # Fool SuperCheckPlugin
                        orig_setUpClass()
                except Exception:
                    self.__exit__(*sys.exc_info())
                    raise

            if orig_setUpClass is klass.__dict__.get('setUpClass', None):
                # was defined on this class, state we wrap it
                setUpClass.__wrapped__ = orig_setUpClass

            def tearDownClass(cls):
                if orig_tearDownClass is not None:
                    super  # Fool SuperCheckPlugin
                    orig_tearDownClass()
                self.__exit__(None, None, None)

            if orig_tearDownClass is klass.__dict__.get('tearDownClass', None):
                # was defined on this class, state we wrap it
                tearDownClass.__wrapped__ = orig_tearDownClass

            klass.setUpClass = classmethod(setUpClass)
            klass.tearDownClass = classmethod(tearDownClass)

            return klass
        else:
            decorated = super(TestCaseContextDecorator, self).__call__(decorable)
            if inspect.isfunction(decorable):
                decorated.__wrapped__ = decorable
            return decorated


class SwitchContextManager(TestCaseContextDecorator):
    """
    Allows temporarily enabling or disabling a switch.

    Ideal for testing.

    >>> @switches(my_switch_name=True)
    >>> def foo():
    >>>     print(gargoyle.is_active('my_switch_name'))

    >>> def foo():
    >>>     with switches(my_switch_name=True):
    >>>         print(gargoyle.is_active('my_switch_name'))

    You may also optionally pass an instance of ``SwitchManager``
    as the first argument.

    >>> def foo():
    >>>     with switches(gargoyle, my_switch_name=True):
    >>>         print(gargoyle.is_active('my_switch_name'))

    Can also wrap unittest classes, which includes Django's
    TestCase classes:

    >>> @switches(my_switch_name=True)
    ... class MyTests(TestCase):
    ...     @classmethod
    ...     def setUpTestData(cls):
    ...         # my_switch_name is True here
    ...
    ...     def test_foo(self):
    ...          # ... and here
    """
    def __init__(self, gargoyle=gargoyle, **keys):
        self.gargoyle = gargoyle
        self.is_active_func = gargoyle.is_active
        self.keys = keys
        self._state = {}
        self._values = {
            True: gargoyle.GLOBAL,
            False: gargoyle.DISABLED,
        }

    def __enter__(self):
        self.patch()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.unpatch()

    def patch(self):
        def is_active(gargoyle):
            is_active_func = gargoyle.is_active

            def wrapped(key, *args, **kwargs):
                if key in self.keys:
                    return self.keys[key]
                return is_active_func(key, *args, **kwargs)
            return wrapped

        self.gargoyle.is_active = is_active(self.gargoyle)

    def unpatch(self):
        self.gargoyle.is_active = self.is_active_func


switches = SwitchContextManager



"""
gargoyle.helpers
~~~~~~~~~~~~~~~~

:copyright: (c) 2010 DISQUS.
:license: Apache License 2.0, see LICENSE for more details.
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import datetime
import json
import uuid

from django.core.serializers.json import DjangoJSONEncoder


class BetterJSONEncoder(DjangoJSONEncoder):
    def default(self, obj):
        if isinstance(obj, uuid.UUID):
            return obj.hex
        elif isinstance(obj, datetime.datetime):
            return obj.strftime('%Y-%m-%dT%H:%M:%S.%fZ')
        elif isinstance(obj, (set, frozenset)):
            return list(obj)
        return super(BetterJSONEncoder, self).default(obj)


def dumps(value, **kwargs):
    return json.dumps(value, cls=BetterJSONEncoder, **kwargs)



"""
gargoyle.templatetags
~~~~~~~~~~~~~~~~~~~~~

:copyright: (c) 2010 DISQUS.
:license: Apache License 2.0, see LICENSE for more details.
"""
from __future__ import absolute_import, division, print_function



from __future__ import absolute_import, division, print_function, unicode_literals

from django.conf import settings
from django.core.cache import caches
from django.http import HttpRequest
from django.utils import six
from django.utils.functional import SimpleLazyObject
from modeldict import ModelDict

from gargoyle.proxy import SwitchProxy

from .constants import DISABLED, EXCLUDE, GLOBAL, INCLUDE, INHERIT, SELECTIVE


class SwitchManager(ModelDict):
    DISABLED = DISABLED
    SELECTIVE = SELECTIVE
    GLOBAL = GLOBAL
    INHERIT = INHERIT

    INCLUDE = INCLUDE
    EXCLUDE = EXCLUDE

    def __init__(self, *args, **kwargs):
        self._registry = {}
        super(SwitchManager, self).__init__(*args, **kwargs)

    def __repr__(self):
        return "<%s: %s (%s)>" % (self.__class__.__name__, self.model, self._registry.values())

    def __getitem__(self, key):
        """
        Returns a SwitchProxy, rather than a Switch. It allows us to
        easily extend the Switches method and automatically include our
        manager instance.
        """
        return SwitchProxy(self, super(SwitchManager, self).__getitem__(key))

    def is_active(self, key, *instances, **kwargs):
        """
        Returns ``True`` if any of ``instances`` match an active switch. Otherwise
        returns ``False``.

        >>> gargoyle.is_active('my_feature', request)
        """
        default = kwargs.pop('default', False)

        # Check all parents for a disabled state
        parts = key.split(':')
        if len(parts) > 1:
            child_kwargs = kwargs.copy()
            child_kwargs['default'] = None
            result = self.is_active(':'.join(parts[:-1]), *instances, **child_kwargs)

            if result is False:
                return result
            elif result is True:
                default = result

        try:
            switch = self[key]
        except KeyError:
            # switch is not defined, defer to parent
            return default

        if switch.status == GLOBAL:
            return True
        elif switch.status == DISABLED:
            return False
        elif switch.status == INHERIT:
            return default

        conditions = switch.value
        # If no conditions are set, we inherit from parents
        if not conditions:
            return default

        if instances:
            # HACK: support request.user by swapping in User instance
            instances = list(instances)
            for v in instances:
                if isinstance(v, HttpRequest) and hasattr(v, 'user'):
                    instances.append(v.user)

        # check each switch to see if it can execute
        return_value = False

        for switch in six.itervalues(self._registry):
            result = switch.has_active_condition(conditions, instances)
            if result is False:
                return False
            elif result is True:
                return_value = True

        # there were no matching conditions, so it must not be enabled
        return return_value

    def register(self, condition_set):
        """
        Registers a condition set with the manager.

        >>> condition_set = MyConditionSet()
        >>> gargoyle.register(condition_set)
        """

        if callable(condition_set):
            registerable = condition_set()
        else:
            registerable = condition_set
        self._registry[registerable.get_id()] = registerable
        return condition_set

    def unregister(self, condition_set):
        """
        Unregisters a condition set with the manager.

        >>> gargoyle.unregister(condition_set)
        """
        if callable(condition_set):
            registerable = condition_set()
        else:
            registerable = condition_set
        popped = self._registry.pop(registerable.get_id(), None)
        return (popped is not None)

    def get_condition_set_by_id(self, switch_id):
        """
        Given the identifier of a condition set (described in
        ConditionSet.get_id()), returns the registered instance.
        """
        return self._registry[switch_id]

    def get_condition_sets(self):
        """
        Returns a generator yielding all currently registered
        ConditionSet instances.
        """
        return six.itervalues(self._registry)

    def get_all_conditions(self):
        """
        Returns a generator which yields groups of lists of conditions.

        >>> for set_id, label, field in gargoyle.get_all_conditions():
        >>>     print("%(label)s: %(field)s" % (label, field.label))
        """
        for condition_set in sorted(self.get_condition_sets(), key=lambda x: x.get_group_label()):
            group = six.text_type(condition_set.get_group_label())
            for field in six.itervalues(condition_set.fields):
                yield condition_set.get_id(), group, field


def make_gargoyle():
    from gargoyle.models import Switch

    kwargs = {
        'key': 'key',
        'value': 'value',
        'instances': True,
        'auto_create': getattr(settings, 'GARGOYLE_AUTO_CREATE', True),
    }

    if hasattr(settings, 'GARGOYLE_CACHE_NAME'):
        kwargs['cache'] = caches[settings.GARGOYLE_CACHE_NAME]

    return SwitchManager(Switch, **kwargs)


gargoyle = SimpleLazyObject(make_gargoyle)



# -*- coding:utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('gargoyle', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='switch',
            name='status',
            field=models.PositiveSmallIntegerField(
                default=1,
                choices=[(1, 'Disabled'), (2, 'Selective'), (3, 'Global'), (4, 'Inherit')],
            ),
        ),
    ]



from __future__ import absolute_import, division, print_function, unicode_literals

import django

# Python 3.2+ includes ContextDecorator, otherwise use backport
try:
    from contextlib import ContextDecorator
except ImportError:
    from contextdecorator import ContextDecorator

# Django 1.9

# url(prefix, include(urls, namespace, name)) -> url(prefix, (urls, namespace, name))
if django.VERSION[:2] >= (1, 9):
    def subinclude(urls_tuple):
        return urls_tuple  # (urls, namespace, name)
else:
    from django.conf.urls import include as subinclude


__all__ = ['ContextDecorator', 'subinclude']



from __future__ import absolute_import, division, print_function, unicode_literals

from django.core.management.base import BaseCommand
from django.utils import six

from gargoyle.models import DISABLED, GLOBAL, Switch


class Command(BaseCommand):
    help = 'Adds or updates the specified gargoyle switch.'

    def add_arguments(self, parser):
        parser.add_argument('switch_name', type=six.text_type)
        parser.add_argument('--disabled', dest='status', action='store_const',
                            default=GLOBAL, const=DISABLED,
                            help='Create a disabled switch.')

    def handle(self, *args, **options):
        switch, created = Switch.objects.get_or_create(
            key=options['switch_name'],
            defaults={
                'status': options['status']
            },
        )
        if not created and switch.status != options['status']:
            switch.status = options['status']
            switch.save()



"""
gargoyle
~~~~~~~~

:copyright: (c) 2010 DISQUS.
:license: Apache License 2.0, see LICENSE for more details.
"""
from __future__ import absolute_import, division, print_function

from django.utils.module_loading import autodiscover_modules

from gargoyle.manager import gargoyle

__version__ = '1.2.7'
VERSION = __version__  # old version compat

__all__ = ('gargoyle', 'autodiscover', '__version__', 'VERSION')

default_app_config = 'gargoyle.apps.GargoyleAppConfig'


def autodiscover():
    """
    Auto-discover INSTALLED_APPS' gargoyle modules and fail silently when
    not present. This forces an import on them to register any gargoyle bits they
    may want.
    """
    import gargoyle.builtins  # noqa
    autodiscover_modules('gargoyle')
